@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import os
 import pathlib
 import pickle
@@ -44,20 +46,25 @@ class ScaleneStatistics:
 
         #   CPU samples for each location in the program
         #   spent in the interpreter
-        self.cpu_samples_python: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.cpu_samples_python: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         #   CPU samples for each location in the program
         #   spent in C / libraries / system calls
-        self.cpu_samples_c: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.cpu_samples_c: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         #   GPU samples for each location in the program
-        self.gpu_samples: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.gpu_samples: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
+
+        #   Number of GPU samples taken (actually weighted by elapsed wallclock time)
+        self.n_gpu_samples: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         #   GPU memory samples for each location in the program
         self.gpu_mem_samples: DefaultDict[
@@ -81,14 +88,14 @@ class ScaleneStatistics:
         self.malloc_samples: Dict[Filename, float] = defaultdict(float)
 
         # malloc samples for each location in the program
-        self.memory_malloc_samples: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.memory_malloc_samples: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         # number of times samples were added for the above
-        self.memory_malloc_count: Dict[
-            Filename, Dict[LineNumber, int]
-        ] = defaultdict(lambda: defaultdict(int))
+        self.memory_malloc_count: Dict[Filename, Dict[LineNumber, int]] = (
+            defaultdict(lambda: defaultdict(int))
+        )
 
         # the current footprint for this line
         self.memory_current_footprint: Dict[
@@ -96,9 +103,9 @@ class ScaleneStatistics:
         ] = defaultdict(lambda: defaultdict(float))
 
         # the max footprint for this line
-        self.memory_max_footprint: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.memory_max_footprint: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         # the current high watermark for this line
         self.memory_current_highwater_mark: Dict[
@@ -118,29 +125,29 @@ class ScaleneStatistics:
         )
 
         # mallocs attributable to Python, for each location in the program
-        self.memory_python_samples: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.memory_python_samples: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         # free samples for each location in the program
-        self.memory_free_samples: Dict[
-            Filename, Dict[LineNumber, float]
-        ] = defaultdict(lambda: defaultdict(float))
+        self.memory_free_samples: Dict[Filename, Dict[LineNumber, float]] = (
+            defaultdict(lambda: defaultdict(float))
+        )
 
         # number of times samples were added for the above
-        self.memory_free_count: Dict[
-            Filename, Dict[LineNumber, int]
-        ] = defaultdict(lambda: defaultdict(int))
+        self.memory_free_count: Dict[Filename, Dict[LineNumber, int]] = (
+            defaultdict(lambda: defaultdict(int))
+        )
 
         # memcpy samples for each location in the program
-        self.memcpy_samples: Dict[
-            Filename, Dict[LineNumber, int]
-        ] = defaultdict(lambda: defaultdict(int))
+        self.memcpy_samples: Dict[Filename, Dict[LineNumber, int]] = (
+            defaultdict(lambda: defaultdict(int))
+        )
 
         # leak score tracking
-        self.leak_score: Dict[
-            Filename, Dict[LineNumber, Tuple[int, int]]
-        ] = defaultdict(lambda: defaultdict(lambda: ((0, 0))))
+        self.leak_score: Dict[Filename, Dict[LineNumber, Tuple[int, int]]] = (
+            defaultdict(lambda: defaultdict(lambda: ((0, 0))))
+        )
 
         self.allocation_velocity: Tuple[float, float] = (0.0, 0.0)
 
@@ -161,7 +168,7 @@ class ScaleneStatistics:
 
         # the peak memory footprint
         self.max_footprint: float = 0.0
-        self.max_footprint_python_fraction : float = 0
+        self.max_footprint_python_fraction: float = 0
         self.max_footprint_loc: Optional[Tuple[Filename, LineNumber]] = None
         # memory footprint samples (time, footprint)
         self.memory_footprint_samples: List[List[float]] = []
@@ -179,9 +186,9 @@ class ScaleneStatistics:
 
         # maps filenames and line numbers to functions (collected at runtime)
         # [filename][lineno] -> function name
-        self.function_map: Dict[
-            Filename, Dict[LineNumber, Filename]
-        ] = defaultdict(lambda: defaultdict(lambda: Filename("")))
+        self.function_map: Dict[Filename, Dict[LineNumber, Filename]] = (
+            defaultdict(lambda: defaultdict(lambda: Filename("")))
+        )
         self.firstline_map: Dict[Filename, LineNumber] = defaultdict(
             lambda: LineNumber(1)
         )
@@ -211,6 +218,7 @@ class ScaleneStatistics:
         self.memcpy_samples.clear()
         self.total_cpu_samples = 0.0
         self.total_gpu_samples = 0.0
+        self.n_gpu_samples.clear()
         self.total_memory_malloc_samples = 0.0
         self.total_memory_free_samples = 0.0
         self.current_footprint = 0.0
@@ -245,12 +253,13 @@ class ScaleneStatistics:
             self.elapsed_time += time.time() - self.start_time
         self.start_time = 0
 
-    def build_function_stats(self, filename: Filename):  # type: ignore
+    def build_function_stats(self, filename: Filename) -> ScaleneStatistics:
         """Produce aggregated statistics for each function."""
         fn_stats = ScaleneStatistics()
         fn_stats.elapsed_time = self.elapsed_time
         fn_stats.total_cpu_samples = self.total_cpu_samples
         fn_stats.total_gpu_samples = self.total_gpu_samples
+        fn_stats.n_gpu_samples = self.n_gpu_samples
         fn_stats.total_memory_malloc_samples = self.total_memory_malloc_samples
         first_line_no = LineNumber(1)
         fn_stats.function_map = self.function_map
@@ -259,6 +268,7 @@ class ScaleneStatistics:
             fn_name = self.function_map[filename][line_no]
             if fn_name == "<module>":
                 continue
+
             fn_stats.cpu_samples_c[fn_name][
                 first_line_no
             ] += self.cpu_samples_c[filename][line_no]
@@ -268,6 +278,9 @@ class ScaleneStatistics:
             fn_stats.gpu_samples[fn_name][first_line_no] += self.gpu_samples[
                 filename
             ][line_no]
+            fn_stats.n_gpu_samples[fn_name][
+                first_line_no
+            ] += self.n_gpu_samples[filename][line_no]
             fn_stats.gpu_mem_samples[fn_name][
                 first_line_no
             ] += self.gpu_mem_samples[filename][line_no]
@@ -315,6 +328,12 @@ class ScaleneStatistics:
             fn_stats.memory_aggregate_footprint[fn_name][
                 first_line_no
             ] += self.memory_aggregate_footprint[filename][line_no]
+
+        for fn_name in fn_stats.gpu_samples:
+            n_gpu_samples = fn_stats.n_gpu_samples[fn_name][first_line_no]
+            if n_gpu_samples != 0:
+                fn_stats.gpu_samples[fn_name][first_line_no] /= n_gpu_samples
+
         return fn_stats
 
     payload_contents = [
@@ -436,6 +455,9 @@ class ScaleneStatistics:
                 self.increment_per_line_samples(
                     self.gpu_samples, x.gpu_samples
                 )
+                self.increment_per_line_samples(
+                    self.n_gpu_samples, x.n_gpu_samples
+                )
                 self.increment_per_line_samples(
                     self.memcpy_samples, x.memcpy_samples
                 )
